@@ -173,9 +173,13 @@ predict.deepregression <- function(
   apply_fun = tfd_mean,
   convert_fun = as.matrix,
   ...
-)
-{
-
+){
+  
+  if(object$engine == "torch" & identical(apply_fun,tfd_mean))
+      apply_fun = function(x) x$mean
+  if(object$engine == "torch" & identical(apply_fun,tfd_stddev))
+    apply_fun = function(x) x$stddev
+    
   # image case
   if(length(object$init_params$image_var)>0 | !is.null(batch_size)){
     
@@ -184,15 +188,46 @@ predict.deepregression <- function(
   }else{
     
     if(is.null(newdata)){
-      yhat <- object$model(prepare_data(object$init_params$parsed_formulas_contents,
-                                        gamdata = object$init_params$gamdata$data_trafos))
+      
+      input_model <- prepare_data(object$init_params$parsed_formulas_contents,
+                                  gamdata = object$init_params$gamdata$data_trafos,
+                                  engine = object$engine)
+      
+      if(object$engine == "torch") {
+        input_model <- 
+          prepare_data_torch(object$init_params$parsed_formulas_contents,
+                             input_x = input_model, object = object)
+        object$model <- object$model()
+        object$model$eval()
+      }
+      if(object$engine == "tf"){
+        yhat <- object$model(input_model, training = FALSE)
+      } else {
+        yhat <- object$model(input_model)
+      }
     }else{
       # preprocess data
       if(is.data.frame(newdata)) newdata <- as.list(newdata)
-      newdata_processed <- prepare_newdata(object$init_params$parsed_formulas_contents, 
-                                           newdata, 
-                                           gamdata = object$init_params$gamdata$data_trafos)
-      yhat <- object$model(newdata_processed)
+      
+      newdata_processed <- prepare_newdata(
+        pfc = object$init_params$parsed_formulas_contents, 
+        newdata = newdata,
+        gamdata = object$init_params$gamdata$data_trafos,
+        engine = object$engine)
+      
+      if(object$engine == "torch") {
+        newdata_processed <- 
+          prepare_data_torch(object$init_params$parsed_formulas_contents,
+                             input_x = newdata_processed, object = object)
+        object$model <- object$model()
+        object$model$eval()
+      }
+      
+      if(object$engine == "tf"){
+        yhat <- object$model(newdata_processed, training = FALSE)
+      } else {
+        yhat <- object$model(newdata_processed)
+      }
     }
   }
   
@@ -218,7 +253,7 @@ fitted.deepregression <- function(
 )
 {
   return(
-    predict.deepregression(object, apply_fun=apply_fun, ...)
+    predict.deepregression(object, apply_fun = apply_fun, ...)
   )
 }
 
@@ -237,14 +272,17 @@ fitted.deepregression <- function(
 #' @param validation_data optional specified validation data
 #' @param validation_split float in [0,1] defining the amount of data used for validation
 #' @param callbacks a list of callbacks for fitting
-#' @param convertfun function to convert R into Tensor object
+#' @param na_handler function to deal with NAs
 #' @param ... further arguments passed to
 #' \code{keras:::fit.keras.engine.training.Model}
 #'
 #'
+#'
 #' @export fit.deepregression
 #' @export
 #' 
+#' @importFrom luz luz_callback_early_stopping luz_callback_keep_best_model
+#' 
 #' @rdname methodDR
 #'
 fit.deepregression <- function(
@@ -260,41 +298,65 @@ fit.deepregression <- function(
   validation_data = NULL,
   validation_split = ifelse(is.null(validation_data), 0.1, 0),
   callbacks = list(),
-  convertfun = function(x) tf$constant(x, dtype="float32"),
+  na_handler = na_omit_list,
   ...
 )
 {
 
   # make callbacks
   if(save_weights){
+    if(object$engine == "tf"){
     weighthistory <- WeightHistory$new()
     callbacks <- append(callbacks, weighthistory)
+    } else {
+      
+      stop("Not implemented yet")
+      
+    }
   }
-  if(early_stopping & length(callbacks)==0)
+  if(early_stopping & length(callbacks)==0){
+    
+    if(object$engine == "tf"){
     callbacks <- append(callbacks,
                         list(callback_terminate_on_naan(),
                              callback_early_stopping(patience = patience,
                                                      restore_best_weights = TRUE,
                                                      monitor = early_stopping_metric)
-                             )
-    )
+                        )
+    )} else {
+      callbacks <- append(callbacks,
+        list(luz::luz_callback_early_stopping(patience = patience),
+             luz::luz_callback_keep_best_model()
+        ))
+    }
+  }
+    
   
   args <- list(...)
 
   input_x <- prepare_data(object$init_params$parsed_formulas_content, 
-                          gamdata = object$init_params$gamdata$data_trafos)
+                          gamdata = object$init_params$gamdata$data_trafos, 
+                          na_handler = na_handler,
+                          engine = object$engine)
   input_y <- as.matrix(object$init_params$y)
+  if(length(attr(input_x, "na_loc"))>0)
+    input_y <- input_y[-attr(input_x, "na_loc"),]
   
-  if(!is.null(validation_data))
+  if(!is.null(validation_data)){
     validation_data <- 
-    list(
-      x = prepare_newdata(object$init_params$parsed_formulas_content, 
-                          validation_data[[1]], 
-                          gamdata = object$init_params$gamdata$data_trafos),
-      y = object$init_params$prepare_y_valdata(validation_data[[2]])
-    )
+        list(
+          x = prepare_newdata(object$init_params$parsed_formulas_content, 
+                              newdata = validation_data[[1]], 
+                              gamdata = object$init_params$gamdata$data_trafos, 
+                              na_handler = na_handler,
+                              engine = object$engine),
+          y = object$init_params$prepare_y_valdata(validation_data[[2]])
+        )
+    if(length(attr(validation_data$x, "na_loc"))>0)
+      validation_data$y <- validation_data$y[-attr(validation_data$x, "na_loc"),]  
+      }
 
-  if(length(object$init_params$image_var)>0){
+  if(length(object$init_params$image_var)>0 & object$engine == "tf"){
     
     args <- prepare_generator_deepregression(
       x = object$model,
@@ -309,33 +371,38 @@ fit.deepregression <- function(
       callbacks = callbacks,
       ...
     )
-
-  }else{
+  } else{
     
-    input_list_model <-
-      list(object = object$model,
-           epochs = epochs,
-           batch_size = batch_size,
-           validation_split = validation_split,
-           validation_data = validation_data,
-           callbacks = callbacks,
-           verbose = verbose,
-           view_metrics = ifelse(view_metrics, getOption("keras.view_metrics", default = "auto"), FALSE)
-      )
+    input_list_model <- 
+      prepare_input_list_model(input_x = input_x,
+                               input_y = input_y,
+                               object = object,
+                               epochs = epochs,
+                               batch_size = batch_size,
+                               validation_split = validation_split,
+                               validation_data = validation_data,
+                               callbacks = callbacks,
+                               verbose = verbose,
+                               view_metrics = view_metrics)
     
-    input_list_model <- c(input_list_model,
-                          list(x = input_x,
-                               y = input_y
-                          ))
     
     args <- append(args,
                    input_list_model[!names(input_list_model) %in%
                                       names(args)])
-    
+    check_input_args_fit(args = args, fit_fun = object$fit_fun)
   }
 
   ret <- suppressWarnings(do.call(object$fit_fun, args))
-  if(save_weights) ret$weighthistory <- weighthistory$weights_last_layer
+  if(save_weights){
+    if(object$engine == "tf"){
+      ret$weighthistory <- weighthistory$weights_last_layer
+      }else {
+        
+        stop("Not implemented yet")
+        
+      }
+    }
+    
   invisible(ret)
 }
 
@@ -373,10 +440,11 @@ coef.deepregression <- function(
   }
   pfc <- pfc[as.logical(to_return)]
   check_names <- names
-  check_names[check_names=="(Intercept)"] <- "1"
+  if(object$engine == "tf") check_names[check_names=="(Intercept)"] <- "1"
   
   coefs <- lapply(1:length(check_names), function(i) 
-    pfc[[i]]$coef(get_weight_by_name(object, check_names[i], which_param)))
+    pfc[[i]]$coef(get_weight_by_name(object, check_names[i], 
+                                     param_nr = which_param)))
   
   names(coefs) <- names
   
@@ -400,7 +468,15 @@ print.deepregression <- function(
   ...
 )
 {
-  print(x$model)
+  
+    if(x$engine == "torch"){
+    subnetworks_index <- which(lapply(strsplit(names(x$model()$modules), "[.]"),
+                   function(x) x[length(x)]) == "subnetwork")
+    amount_params <- seq_len(length(subnetworks_index))
+    model_summary <- x$model()$modules[subnetworks_index]
+    names(model_summary) <- names(x$init_params$additive_predictors)[amount_params]
+    print(model_summary)
+    } else print(x$model)
   fae <- x$init_params$list_of_formulas
   cat("Model formulas:\n---------------\n")
   invisible(sapply(1:length(fae), function(i){ cat(names(fae)[i],":\n"); print(fae[[i]])}))
@@ -461,9 +537,13 @@ cv.deepregression <- function(
       data_size = NROW(x$init_params$y),
       cv_folds)
   }
+  if(x$engine ==  "tf") {
+    old_weights <- x$model$get_weights()
+  } else {
+    old_weights <- x %>% get_weights_torch()
+    }
+  
   
-  old_weights <- x$model$get_weights()
-
   # subset fun
   if(NCOL(x$init_params$y)==1)
     subset_fun <- function(y,ind) y[ind] else
@@ -476,55 +556,79 @@ cv.deepregression <- function(
     if(print_folds) cat("Fitting Fold ", folds_iter, " ... ")
     st1 <- Sys.time()
 
-    # does not work?
-    # this_mod <- clone_model(x$model)
-    this_mod <- x$model
+    this_mod <- x
 
     train_ind <- this_fold[[1]]
     test_ind <- this_fold[[2]]
-
-    x_train <- prepare_data(x$init_params$parsed_formulas_content,
-                            gamdata = x$init_params$gamdata$data_trafos)
+    
+    x_train <- prepare_data(pfc = x$init_params$parsed_formulas_content,
+                            gamdata = x$init_params$gamdata$data_trafos,
+                            engine = x$engine)
     
     train_data <- lapply(x_train, function(x)
         subset_array(x, train_ind))
     test_data <- lapply(x_train, function(x)
         subset_array(x, test_ind))
     
-    # make callbacks
+    
     this_callbacks <- callbacks
     if(save_weights){
-      weighthistory <- WeightHistory$new()
-      this_callbacks <- append(this_callbacks, weighthistory)
+      if(x$engine=="tf"){
+        weighthistory <- WeightHistory$new()
+        this_callbacks <- append(this_callbacks, weighthistory)
+        } else {
+          
+          stop("Not implemented yet.")
+        
+          }
     }
 
     args <- list(...)
+    
+    input_list_model <- prepare_input_list_model(
+                          input_x = train_data,
+                          input_y = subset_fun(x$init_params$y, train_ind),
+                          object = this_mod, 
+                          validation_split = NULL,
+                           validation_data = list(
+                             test_data,
+                             subset_fun(x$init_params$y,test_ind)
+                           ),
+                           callbacks = this_callbacks,
+                           verbose = verbose,
+                           view_metrics = FALSE)
+    
     args <- append(args,
-                   list(object = this_mod,
-                        x = train_data,
-                        y = subset_fun(x$init_params$y, train_ind),
-                        validation_split = NULL,
-                        validation_data = list(
-                          test_data,
-                          subset_fun(x$init_params$y,test_ind)
-                        ),
-                        callbacks = this_callbacks,
-                        verbose = verbose,
-                        view_metrics = FALSE
-                   )
-    )
+                   input_list_model[!names(input_list_model) %in%
+                                      names(args)])
     
     args <- append(args, x$init_params$ellipsis)
 
     ret <- do.call(x$fit_fun, args)
-    if(save_weights) ret$weighthistory <- weighthistory$weights_last_layer
-    if(!is.null(save_fun))
-      ret$save_fun_result <- save_fun(this_mod)
+    if(save_weights){
+      if(x$engine=="tf"){
+        ret$weighthistory <- weighthistory$weights_last_layer
+      } else {
+          
+        stop("Not implemented yet.")
+        
+        }
+      }
+    
+    if(!is.null(save_fun)){
+      #ret$save_fun_result <- save_fun(x, train_ind, test_ind)
+      ret$save_fun_result <- save_fun(this_mod, train_ind, test_ind)
+    }
     
     if(stop_if_nan && any(is.nan(ret$metrics$validloss)))
       stop("Fold ", folds_iter, " with NaN's in ")
+    
+    if(x$engine == "tf") {
+      this_mod$model$set_weights(old_weights)
+    } else {
+      this_mod$model()$load_state_dict(old_weights)
+      }
 
-    this_mod$set_weights(old_weights)
     td <- Sys.time()-st1
     if(print_folds) cat("\nDone in", as.numeric(td), "", attr(td,"units"), "\n")
 
@@ -534,10 +638,15 @@ cv.deepregression <- function(
 
   class(res) <- c("drCV","list")
 
-  if(plot) try(plot_cv(res), silent = TRUE)
-
-  x$model$set_weights(old_weights)
+  if(plot) try(plot_cv(res, engine = x$engine), silent = TRUE)
 
+  # CM:why is this needed? We don't use x in the cv
+  if(x$engine == "tf") {
+    x$model$set_weights(old_weights)
+  } else {
+    x$model()$load_state_dict(old_weights)
+  }
+  
   invisible(return(res))
 
 }
@@ -560,7 +669,8 @@ mean.deepregression <- function(
   ...
 )
 {
-  predict.deepregression(x, newdata = data, apply_fun = tfd_mean, ...)
+  
+  predict.deepregression(x, apply_fun = tfd_mean, newdata = data, ...)
 }
 
 
@@ -619,9 +729,14 @@ quant.deepregression <- function(
   ...
 )
 {
+  
+  if(x$engine == 'tf') apply_fun = function(x) tfd_quantile(x, value=probs)
+  if(x$engine == 'torch') apply_fun = function(x) x$icdf(value = probs)
+  
+  
   predict.deepregression(x,
                          newdata = data,
-                         apply_fun = function(x) tfd_quantile(x, value=probs),
+                         apply_fun = apply_fun,
                          ...)
 }
 
@@ -640,14 +755,29 @@ get_distribution <- function(
 )
 {
   if(is.null(data)){
-    disthat <- x$model(prepare_data(x$init_params$parsed_formulas_content, 
-                                    gamdata = x$init_params$gamdata$data_trafos))
+    model_input <- prepare_data(x$init_params$parsed_formulas_content, 
+                                gamdata = x$init_params$gamdata$data_trafos,
+                                engine = x$engine)
+    if(x$engine == "torch"){
+      model_input <- 
+          prepare_data_torch(x$init_params$parsed_formulas_contents,
+                             input_x = model_input, object = x)
+        x$model <- x$model()
+    }
+    disthat <- x$model(model_input)
   }else{
     # preprocess data
     if(is.data.frame(data)) data <- as.list(data)
     newdata_processed <- prepare_newdata(x$init_params$parsed_formulas_content, 
-                                         data, 
-                                         gamdata = x$init_params$gamdata$data_trafos)
+                                         newdata = data, 
+                                         gamdata = x$init_params$gamdata$data_trafos,
+                                         engine = x$engine)
+    if(x$engine == "torch"){
+      newdata_processed <- 
+        prepare_data_torch(x$init_params$parsed_formulas_contents,
+                           input_x = newdata_processed, object = x)
+      x$model <- x$model()
+    }
     disthat <- x$model(newdata_processed)
   }
   return(disthat)
@@ -668,30 +798,55 @@ log_score <- function(
   x,
   data=NULL,
   this_y=NULL,
-  ind_fun = function(x) tfd_independent(x),
+  ind_fun = NULL,
   convert_fun = as.matrix,
   summary_fun = function(x) x
 )
 {
 
+  if(x$engine == "tf") {
+    ind_fun <- function(x) tfd_independent(x)
+    log_prob <- function(x, value) tfd_log_prob(x, value)
+    } else {
+      ind_fun <- function(x) x
+      log_prob <- function(x, value) x$log_prob(value)
+      }
+  
+  
+  
   if(is.null(data)){
     
     this_data <- prepare_data(x$init_params$parsed_formulas_content, 
-                              gamdata = x$init_params$gamdata$data_trafos)
+                              gamdata = x$init_params$gamdata$data_trafos,
+                              engine = x$engine)
+    if(x$engine == "torch"){
+      this_data <- prepare_data_torch(pfc = x$init_params$parsed_formulas_content,
+                         input_x = this_data, object = x)
+      x$model <- x$model()
+    }
   
   }else{
     
     if(is.data.frame(data)) data <- as.list(data)
+    
     this_data <- prepare_newdata(x$init_params$parsed_formulas_content, 
                                  data, 
-                                 gamdata = x$init_params$gamdata$data_trafos)
+                                 gamdata = x$init_params$gamdata$data_trafos,
+                                 engine = x$engine)
+    
+    if(x$engine == "torch"){
+      this_data <- prepare_data_torch(pfc = x$init_params$parsed_formulas_content,
+                                      input_x = this_data, object = x)
+      x$model <- x$model()
+      x$model$eval()
+    }
     
   }
   
   disthat <- x$model(this_data)
     
   if(is.null(this_y)){
-    this_y <- x$init_params$y
+    this_y <- as.matrix(x$init_params$y)
   }else{
     if(is.null(dim(this_y))){
       warning("Meaningful log-score calculations require this_y to be a matrix (forced now).")
@@ -700,7 +855,7 @@ log_score <- function(
   }
   
   return(summary_fun(convert_fun(
-    disthat %>% ind_fun() %>% tfd_log_prob(this_y)
+    disthat %>% ind_fun() %>% log_prob(this_y)
   )))
 }
 
@@ -719,12 +874,17 @@ get_weight_by_name <- function(mod, name, param_nr=1, postfixes="")
 
   # check for shared layer  
   names_pfc <- get_names_pfc(mod$init_params$parsed_formulas_contents[[param_nr]])
-  names_pfc[names_pfc=="(Intercept)"] <- "1"
+  if(mod$engine == "tf") names_pfc[names_pfc=="(Intercept)"] <- "1"
   pfc_term <- mod$init_params$parsed_formulas_contents[[param_nr]][[which(names_pfc==name)]]
   if(!is.null(pfc_term$shared_name)){
     this_name <- paste0(pfc_term$shared_name, postfixes)
   }else{
-    this_name <- paste0(makelayername(name, param_nr), postfixes)
+    
+    if(mod$engine == "tf"){
+      this_name <- paste0(makelayername(name, param_nr), postfixes)
+      } else {
+        this_name <- paste(name, param_nr, sep = "_")
+        }
   }
   # names <- get_mod_names(mod)
   if(length(this_name)>1){
