@@ -20,9 +20,9 @@ test_that("Simple additive model", {
       list_of_deep_models = list(d = deep_model)
     )
     expect_is(mod, "deepregression")
-    expect_length(mod, 3)
+    expect_length(mod, 4)
     expect_true(length(setdiff(names(mod), 
-                               c("model", "init_params", "fit_fun")
+                               c("model", "init_params", "fit_fun", "engine")
                                )
                        )==0)
   }
@@ -38,9 +38,9 @@ test_that("Simple additive model", {
   )
   
   expect_is(mod, "deepregression")
-  expect_length(mod, 3)
+  expect_length(mod, 4)
   expect_true(length(setdiff(names(mod), 
-                             c("model", "init_params", "fit_fun")
+                             c("model", "init_params", "fit_fun", "engine")
   )
   )==0)
 
@@ -53,9 +53,9 @@ test_that("Simple additive model", {
     list_of_deep_models = list(d = deep_model, g = deep_model)
   )
   expect_is(mod, "deepregression")
-  expect_length(mod, 3)
+  expect_length(mod, 4)
   expect_true(length(setdiff(names(mod), 
-                             c("model", "init_params", "fit_fun")
+                             c("model", "init_params", "fit_fun", "engine")
   )
   )==0)
   
